#include "matrix_operation.h"
#include "vector.h"
#include <stdlib.h>
#include <stdio.h>
#include "sparse_matrix.h"

void multiply_vector_matrix(const double *vector, const SparseMatrix *matrix, double *result) {
    init_vector(result, matrix->num_nodes, 0.0);
    for (int i = 0; i < matrix->num_arcs; ++i) {
        int origin = matrix->arcs[i].origin;
        int dest = matrix->arcs[i].dest;
        result[dest] += vector[origin] * matrix->arcs[i].value;
    }
}


void multiply_vector_matrix_parallel(const double *vector, const SparseMatrix *matrix, double *result) {
    int num_nodes = matrix->num_nodes;
    int num_arcs = matrix->num_arcs;
    init_vector(result, num_nodes, 0.0);

    #pragma omp parallel
    {
        double *local_result = (double *)calloc(num_nodes, sizeof(double));

        // parallelize for loop
        #pragma omp for
        for (int i = 0; i < num_arcs; ++i) {
            int origin = matrix->arcs[i].origin;
            int dest = matrix->arcs[i].dest;
            local_result[dest] += vector[origin] * matrix->arcs[i].value;
        }

        // merge results
        for (int j = 0; j < num_nodes; ++j) {
            double val = local_result[j];
            if (val != 0.0) {
                #pragma omp atomic
                result[j] += val;
            }
        }

        free(local_result);
    }
}

void convert_to_stochastic(SparseMatrix *matrix) {
    double *non_zero = (double *)calloc(matrix->num_nodes, sizeof(double));
    if (!non_zero) {
        fprintf(stderr, "Memory allocation failed\n");
        return;
    }

    // count non zero values
    for (int i = 0; i < matrix->num_arcs; i++) {
        non_zero[matrix->arcs[i].dest] += matrix->arcs[i].value;
    }

    // normalize values
    for (int i = 0; i < matrix->num_arcs; i++) {
        int dest = matrix->arcs[i].dest;
        double old_value = matrix->arcs[i].value;
        matrix->arcs[i].value = old_value / (double)non_zero[dest];
    }

    free(non_zero);
}

int compare_arcs(const void *a, const void *b) {
    Arc *arc_a = (Arc *)a;
    Arc *arc_b = (Arc *)b;
    return arc_a->origin - arc_b->origin;
}

